# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# generated by datamodel-codegen:
#   filename:  rms_openapi.json
#   timestamp: 2024-02-05T12:33:03+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from typing_extensions import Literal


class ApplicationInfo(BaseModel):
    name: str = Field(..., description="Application name.", title="Name")
    version: str = Field(..., description="Application version.", title="Version")
    install_path: str = Field(
        ..., description="Installation path of application.", title="Install Path"
    )
    executable: str = Field(
        ..., description="Executable path to run application.", title="Executable"
    )
    environment: Optional[Dict[str, Any]] = Field(
        None, description="Environment setup for the process.", title="Environment"
    )
    capabilities: Optional[List[str]] = Field(
        None, description="Capabilities of the application.", title="Capabilities"
    )
    customization_hook: Optional[Dict[str, Any]] = Field(
        None,
        description="Details of a custom hook used to modify the configuration before runs are performed.",
        title="Customization Hook",
    )


class EvaluatorTaskDirectoryCleanup(Enum):
    always = "always"
    on_success = "on_success"
    never = "never"


class ContextUpdate(BaseModel):
    custom: Optional[Dict[str, Optional[Union[int, bool, str, float]]]] = Field(
        {}, description="Custom runtime properties.", title="Custom"
    )


class CrsCountResponse(BaseModel):
    num_compute_resource_sets: Optional[int] = Field(0, title="Num Compute Resource Sets")


class TaskDirectoryCleanupEnum(Enum):
    always = "always"
    on_success = "on_success"
    never = "never"


class EvaluatorRegistration(BaseModel):
    id: Optional[str] = Field(None, description="Unique ID for this worker.", title="Id")
    name: Optional[str] = Field(
        None, description="User-defined name for this worker.", title="Name"
    )
    last_modified: Optional[datetime] = Field(
        None,
        description="Date and time when the registration was last modified.",
        title="Last Modified",
    )
    host_id: Optional[str] = Field(
        None, description="Static, hardware and configuration-based UUID.", title="Host Id"
    )
    host_name: Optional[str] = Field(
        None, description="Name of the host that the worker is running on.", title="Host Name"
    )
    username: Optional[str] = Field(
        None, description="Username that the worker authenticated with.", title="Username"
    )
    platform: Optional[str] = Field(
        None, description="OS that the evaluator is running on.", title="Platform"
    )
    build_info: Optional[Dict[str, Any]] = Field(
        {}, description="Detailed build information.", title="Build Info"
    )
    compute_resource_set_id: Optional[str] = Field(
        None,
        description="ID of the compute resource set that the evaluator belongs to.",
        title="Compute Resource Set Id",
    )
    change_requested: Optional[datetime] = Field(
        None,
        description="Date and time of the configuration's last modification request.",
        title="Change Requested",
    )


class EvaluatorsCountResponse(BaseModel):
    num_evaluators: Optional[int] = Field(0, title="Num Evaluators")


class EvaluatorsRequest(BaseModel):
    evaluators: List[EvaluatorRegistration] = Field(
        ..., description="Evaluator details", title="Evaluators"
    )


class EvaluatorsResponse(BaseModel):
    evaluators: List[EvaluatorRegistration] = Field(
        ..., description="Evaluator details", title="Evaluators"
    )


class HpcResources(BaseModel):
    num_cores_per_node: Optional[int] = Field(
        None, description="Number of cores per node.", title="Num Cores Per Node"
    )
    num_gpus_per_node: Optional[int] = Field(
        None, description="Number of GPUs per node.", title="Num Gpus Per Node"
    )
    exclusive: Optional[bool] = Field(
        None, description="To not share nodes with other running jobs.", title="Exclusive"
    )
    queue: Optional[str] = Field(None, description="Scheduler's queue.", title="Queue")


class KubernetesKedaBackend(BaseModel):
    plugin_name: Literal["kubernetes"] = Field(..., title="Plugin Name")
    debug: Optional[bool] = Field(
        False, description="Enable additional debugging of the backend", title="Debug"
    )
    job_script_template_path: Optional[str] = Field(
        None,
        description="Path to the job script template to use in the backend.",
        title="Job Script Template Path",
    )
    working_dir: Optional[str] = Field(
        None, description="Working directory to use in the backend.", title="Working Dir"
    )
    env: Optional[Dict[str, Any]] = Field(
        None, description="Static environment variables needed for job execution.", title="Env"
    )
    cpu_limit: Optional[str] = Field(
        "1.0", description="CPU limit applied to each evaluator instance.", title="Cpu Limit"
    )
    memory_limit: Optional[str] = Field(
        "250M", description="Memory limit applied to each evaluator instance.", title="Memory Limit"
    )
    namespace: Optional[str] = Field(
        "default", description="Kubernetes namespace to use to scale evaluators.", title="Namespace"
    )
    target_resource_kind: Optional[str] = Field(
        "job",
        description="Kubernetes resource kind that the REP scaler should target. Options are ``job``, ``deployment`` and ``statefulset``.",
        title="Target Resource Kind",
    )


class KubernetesResourceScaling(BaseModel):
    plugin_name: Literal["kubernetes_resource_scaling"] = Field(..., title="Plugin Name")
    target_resource_kind: Optional[str] = Field(
        "job",
        description="Kubernetes resource kind that the REP scaler should target. Options are ``job``, ``deployment`` and ``statefulset``.",
        title="Target Resource Kind",
    )


class LocalBackend(BaseModel):
    plugin_name: Literal["local"] = Field(..., title="Plugin Name")
    debug: Optional[bool] = Field(
        False, description="Enable additional debugging of the backend", title="Debug"
    )
    working_dir: Optional[str] = Field(
        None, description="Working directory to use in the backend.", title="Working Dir"
    )
    env: Optional[Dict[str, Any]] = Field(
        None, description="Static environment variables needed for job execution.", title="Env"
    )


class Machine(BaseModel):
    name: str = Field(..., description="Name of the machine", title="Name")
    num_cores: int = Field(..., description="Number of cores available", title="Num Cores")


class MaxAvailableResourceScaling(BaseModel):
    plugin_name: Literal["max_available_resource_scaling"] = Field(..., title="Plugin Name")
    match_all_requirements: Optional[bool] = Field(
        False,
        description="Whether scaling should work with available resource properties specified in compute resource set (default) or require a match of all requirements of the task definition.",
        title="Match All Requirements",
    )


class MockupBackend(BaseModel):
    plugin_name: Literal["mockup"] = Field(..., title="Plugin Name")
    debug: Optional[bool] = Field(
        False, description="Enable additional debugging of the backend", title="Debug"
    )


class Node(BaseModel):
    name: Optional[str] = Field(None, description="Node name", title="Name")
    total_memory_mb: Optional[int] = Field(
        ..., description="Total memory.", title="Total Memory Mb"
    )
    total_cores: Optional[int] = Field(..., description="Number of cores.", title="Total Cores")
    additional_props: Optional[Dict[str, Any]] = Field({}, title="Additional Props")


class NodeGroup(BaseModel):
    node_names: List[str] = Field(..., title="Node Names")
    memory_per_node_mb: Optional[int] = Field(
        ..., description="Memory per node.", title="Memory Per Node Mb"
    )
    cores_per_node: Optional[int] = Field(
        ..., description="Cores per node.", title="Cores Per Node"
    )


class PlatformEnum(Enum):
    windows = "windows"
    linux = "linux"
    darwin = "darwin"


class ProblemDetail(BaseModel):
    type: Optional[str] = Field(None, title="Type")
    title: Optional[str] = Field(None, title="Title")
    status: int = Field(..., title="Status")
    detail: str = Field(..., title="Detail")
    instance: Optional[str] = Field(None, title="Instance")


class ProcessLauncherProcessRunner(BaseModel):
    plugin_name: Literal["process_launcher_module"] = Field(..., title="Plugin Name")
    default_user: Optional[str] = Field(
        None, description="User to use when none is specified.", title="Default User"
    )
    timeout: Optional[int] = Field(
        30, description="Timeout in seconds before the request is aborted.", title="Timeout"
    )
    allowed_users: Optional[List[str]] = Field(
        None, description="Users allowed to launch processes.", title="Allowed Users"
    )
    disallowed_users: Optional[List[str]] = Field(
        ["root"], description="Users not allowed to launch processes.", title="Disallowed Users"
    )
    user_mapping: Optional[Dict[str, str]] = Field(
        {}, description="Map of calling user to system user.", title="User Mapping"
    )
    minimum_uid: Optional[int] = Field(
        1000, description="Minimum UID of users allowed to launch processes.", title="Minimum Uid"
    )
    minimum_gid: Optional[int] = Field(
        1000, description="Minimum GID of users allowed to launch processes.", title="Minimum Gid"
    )


class Queue(BaseModel):
    name: Optional[str] = Field(None, description="Queue name", title="Name")
    node_groups: Optional[List[NodeGroup]] = Field(
        None,
        description="List of node groups associated with the queue (if available).",
        title="Node Groups",
    )
    additional_props: Optional[Dict[str, Any]] = Field({}, title="Additional Props")


class Resources(BaseModel):
    num_cores: Optional[int] = Field(None, description="Number of cores.", title="Num Cores")
    platform: Optional[PlatformEnum] = Field(
        None, description="Basic platform information. Options are ``'linux'`` and ``'windows'``."
    )
    memory: Optional[int] = Field(None, description="Amount of RAM in bytes.", title="Memory")
    disk_space: Optional[int] = Field(
        None, description="Amount of disk space in bytes.", title="Disk Space"
    )
    custom: Optional[Dict[str, Optional[Union[bool, int, str, float]]]] = Field(
        {}, description="Custom resource properties.", title="Custom"
    )
    num_instances: Optional[int] = Field(
        None,
        description="Number of instances/jobs that can be created on the compute resource set.",
        title="Num Instances",
    )


class RestLauncherProcessRunner(BaseModel):
    plugin_name: Literal["process_launcher_service"] = Field(..., title="Plugin Name")
    launcher_url: Optional[str] = Field(
        "http://localhost:4911",
        description="URL to use when none is specified.",
        title="Launcher Url",
    )
    verify_ssl: Optional[bool] = Field(
        True, description="Check the SSL certificate for HTTPS launchers.", title="Verify Ssl"
    )
    timeout: Optional[int] = Field(
        30, description="Timeout in seconds before the request is aborted.", title="Timeout"
    )
    shell: Optional[bool] = Field(
        True, description="Enable the shell interpretation on subprocess run.", title="Shell"
    )


class ScalerApplicationInfo(BaseModel):
    name: str = Field(..., description="Application name.", title="Name")
    version: str = Field(..., description="Application version.", title="Version")
    install_path: str = Field(
        ..., description="Installation path of application.", title="Install Path"
    )
    executable: str = Field(
        ..., description="Executable path to run application.", title="Executable"
    )
    environment: Optional[Dict[str, Any]] = Field(
        None, description="Environment setup for the process.", title="Environment"
    )
    capabilities: Optional[List[str]] = Field(
        None, description="Capabilities of the application.", title="Capabilities"
    )
    customization_hook: Optional[Dict[str, Any]] = Field(
        None,
        description="Details of a custom hook used to modify the configuration before runs are performed.",
        title="Customization Hook",
    )
    resource_name: Optional[str] = Field(
        None,
        description="Kubernetes object (deployment/statefulset) name or solver image used as target resource by KEDA",
        title="Resource Name",
    )
    evaluator_image: Optional[str] = Field(
        None, description="Evaluator image to be used", title="Evaluator Image"
    )
    scaling_max_eval_instances: Optional[int] = Field(
        1,
        description="Maximum number of instances that can be created when scaling up.",
        title="Scaling Max Eval Instances",
    )
    scaling_min_eval_instances: Optional[int] = Field(
        0,
        description="Minimum number of instances than can be terminated when scaling down.",
        title="Scaling Min Eval Instances",
    )
    scaling_threshold: Optional[int] = Field(
        1,
        description="Threshold value to determine when Kubernetes deployments should be scaled up or down.",
        title="Scaling Threshold",
    )
    cool_down_period: Optional[int] = Field(
        60,
        description="Period to wait before scaling down the resource to 0 instances.",
        title="Cool Down Period",
    )


class ScalerRegistration(BaseModel):
    id: Optional[str] = Field(None, description="Unique ID for this worker.", title="Id")
    name: Optional[str] = Field(
        None, description="User-defined name for this worker.", title="Name"
    )
    last_modified: Optional[datetime] = Field(
        None,
        description="Date and time when the registration was last modified.",
        title="Last Modified",
    )
    host_id: Optional[str] = Field(
        None, description="Static, hardware and configuration-based UUID.", title="Host Id"
    )
    host_name: Optional[str] = Field(
        None, description="Name of the host that the worker is running on.", title="Host Name"
    )
    username: Optional[str] = Field(
        None, description="Username that the worker authenticated with.", title="Username"
    )
    platform: Optional[str] = Field(
        None, description="OS that the evaluator is running on.", title="Platform"
    )
    build_info: Optional[Dict[str, Any]] = Field(
        {}, description="Detailed build information.", title="Build Info"
    )
    config_modified: Optional[datetime] = Field(
        None,
        description="Date and time of the configuration's last modification.",
        title="Config Modified",
    )


class ScalersCountResponse(BaseModel):
    num_scalers: Optional[int] = Field(0, title="Num Scalers")


class ScalersRequest(BaseModel):
    scalers: List[ScalerRegistration] = Field(..., description="Scaler details", title="Scalers")


class ScalersResponse(BaseModel):
    scalers: List[ScalerRegistration] = Field(..., description="Scaler details", title="Scalers")


class ServiceUserProcessRunner(BaseModel):
    plugin_name: Literal["service_user_module"] = Field(..., title="Plugin Name")


class Status(BaseModel):
    time: str = Field(..., title="Time")
    build: Dict[str, Any] = Field(..., title="Build")


class ClusterInfo(BaseModel):
    id: Optional[str] = Field(None, description="Unique ID for the database.", title="Id")
    crs_id: Optional[str] = Field(None, description="Compute resource set ID.", title="Crs Id")
    name: Optional[str] = Field(None, description="Cluster name.", title="Name")
    queues: Optional[List[Queue]] = Field([], title="Queues")
    nodes: Optional[List[Node]] = Field([], title="Nodes")
    additional_props: Optional[Dict[str, Dict[str, Any]]] = Field({}, title="Additional Props")


class Context(BaseModel):
    custom: Optional[Dict[str, Optional[Union[int, bool, str, float]]]] = Field(
        {}, description="Custom runtime properties.", title="Custom"
    )
    machines_list: Optional[List[Machine]] = Field(
        None,
        description="List of machines for distributed parallel processing.",
        title="Machines List",
    )


class EvaluatorResources(BaseModel):
    num_cores: Optional[int] = Field(None, description="Number of cores.", title="Num Cores")
    platform: Optional[PlatformEnum] = Field(
        None, description="Basic platform information. Options are ``'linux'`` and ``'windows'``."
    )
    memory: Optional[int] = Field(None, description="Amount of RAM in bytes.", title="Memory")
    disk_space: Optional[int] = Field(
        None, description="Amount of disk space in bytes.", title="Disk Space"
    )
    custom: Optional[Dict[str, Optional[Union[bool, int, str, float]]]] = Field(
        {}, description="Custom resource properties.", title="Custom"
    )
    hpc_resources: Optional[HpcResources] = None


class OrchestrationInterfacesBackend(BaseModel):
    plugin_name: Literal["orchestration_interfaces"] = Field(..., title="Plugin Name")
    debug: Optional[bool] = Field(
        False, description="Enable additional debugging of the backend", title="Debug"
    )
    scheduler_type: Optional[str] = Field(
        "slurm",
        description="Job scheduler type, such as ``slurm``, ``pbs``, ``uge``, or ``lsf`` to use in the backend.",
        title="Scheduler Type",
    )
    scheduler_queue_default: Optional[str] = Field(
        None,
        description="Job scheduler queue to use for submission.",
        title="Scheduler Queue Default",
    )
    scheduler_command_override: Optional[str] = Field(
        None,
        description="Path to the JSON file with custom scheduler command definitions.",
        title="Scheduler Command Override",
    )
    scheduler_script_override: Optional[str] = Field(
        None,
        description="Path to the shell script to template for the scheduler.",
        title="Scheduler Script Override",
    )
    exclusive_default: Optional[bool] = Field(
        False,
        description="Request the scheduler to hold the nodes exclusively for one request.",
        title="Exclusive Default",
    )
    distributed_default: Optional[bool] = Field(
        True,
        description="Allow the scheduler to provide multiple machines to fulfill the request.",
        title="Distributed Default",
    )
    num_cores_default: Optional[int] = Field(
        1,
        description="Number of cores to request from the scheduler for a task.",
        title="Num Cores Default",
    )
    working_dir: Optional[str] = Field(
        None, description="Working directory to use in the backend.", title="Working Dir"
    )
    env: Optional[Dict[str, Any]] = Field(
        None, description="Static environment variables needed for job execution.", title="Env"
    )
    process_runner: Optional[
        Union[ServiceUserProcessRunner, ProcessLauncherProcessRunner, RestLauncherProcessRunner]
    ] = Field(
        {"plugin_name": "service_user_module"},
        description="Process runner used to execute commands.",
        discriminator="plugin_name",
        title="Process Runner",
    )
    create_workdir: Optional[bool] = Field(
        True,
        description="Create base and/or user-specific working directories at runtime.",
        title="Create Workdir",
    )
    use_templates: Optional[bool] = Field(
        True,
        description="Use the templated versions of the scripts and write them to the working directory.",
        title="Use Templates",
    )


class ComputeResourceSet(BaseModel):
    name: Optional[str] = Field(
        "default", description="Name of the compute resource set.", title="Name"
    )
    id: Optional[str] = Field(None, description="ID for this set.", title="Id")
    scaler_id: Optional[str] = Field(
        None,
        description="Temporary. To be removed after transitioning to client_id.",
        title="Scaler Id",
    )
    last_modified: Optional[datetime] = Field(
        None, description="Last modified time.", title="Last Modified"
    )
    backend: Optional[
        Union[KubernetesKedaBackend, OrchestrationInterfacesBackend, LocalBackend, MockupBackend]
    ] = Field(
        {"debug": False, "plugin_name": "local"},
        description="Backend to use in this compute resource set.",
        discriminator="plugin_name",
        title="Backend",
    )
    scaling_strategy: Optional[
        Union[MaxAvailableResourceScaling, KubernetesResourceScaling]
    ] = Field(
        {"match_all_requirements": False, "plugin_name": "max_available_resource_scaling"},
        description="Scaling strategy to use in this compute resource set.",
        discriminator="plugin_name",
        title="Scaling Strategy",
    )
    available_resources: Optional[Resources] = Field(
        {"custom": {}}, description="Available resources in the compute resource set."
    )
    available_applications: Optional[List[ScalerApplicationInfo]] = Field(
        [], description="List of available applications.", title="Available Applications"
    )
    evaluator_requirements_matching: Optional[bool] = Field(
        False,
        description="Whether the evaluators should do matching of resource and software requirements.",
        title="Evaluator Requirements Matching",
    )
    evaluator_task_directory_cleanup: Optional[EvaluatorTaskDirectoryCleanup] = Field(
        "always",
        description="Cleanup policy for task directories that are passed to evaluators.",
        title="Evaluator Task Directory Cleanup",
    )
    evaluator_auto_shutdown_time: Optional[int] = Field(
        20,
        description="Time after which to shutdown the evaluator if not running any jobs.",
        title="Evaluator Auto Shutdown Time",
    )
    evaluator_loop_interval: Optional[int] = Field(
        5,
        description="Number of seconds between each iteration of the evaluator's main loop.",
        title="Evaluator Loop Interval",
    )


class ComputeResourceSetsRequest(BaseModel):
    compute_resource_sets: List[ComputeResourceSet] = Field(
        ..., description="Compute resource set details", title="Compute Resource Sets"
    )


class ComputeResourceSetsResponse(BaseModel):
    compute_resource_sets: List[ComputeResourceSet] = Field(
        ..., description="Compute resource set details", title="Compute Resource Sets"
    )


class EvaluatorConfiguration(BaseModel):
    id: Optional[str] = Field(None, description="Unique database ID (read-only).", title="Id")
    evaluator_id: Optional[str] = Field(
        None, description="ID of the parent evaluator (read-only).", title="Evaluator Id"
    )
    last_modified: Optional[datetime] = Field(
        None, description="Last modified time", title="Last Modified"
    )
    working_directory: Optional[str] = Field(None, title="Working Directory")
    local_file_cache_max_size: Optional[int] = Field(
        None, description="Maximum allowed cache size in bytes.", title="Local File Cache Max Size"
    )
    max_num_parallel_tasks: Optional[int] = Field(None, title="Max Num Parallel Tasks")
    task_directory_cleanup: Optional[TaskDirectoryCleanupEnum] = Field(
        None, title="Task Directory Cleanup"
    )
    resources: Optional[EvaluatorResources] = {"custom": {}}
    task_manager_type: Optional[str] = Field(None, title="Task Manager Type")
    loop_interval: Optional[float] = Field(
        5.0,
        description="Number of seconds between each iteration of the evaluator's main loop.",
        title="Loop Interval",
    )
    local_file_cache: Optional[bool] = Field(
        True,
        description="Whether to configure a local file cache in the file tool.",
        title="Local File Cache",
    )
    applications: Optional[List[ApplicationInfo]] = Field(
        [], description="List of available applications.", title="Applications"
    )
    project_server_select: Optional[bool] = Field(
        True,
        description="Get project assignments from the server instead of using the locally set values",
        title="Project Server Select",
    )
    project_list: Optional[List[str]] = Field(
        [],
        description="IDs of projects that the evaluator should work on, in order",
        title="Project List",
    )
    project_assignment_mode: Optional[str] = Field(
        "all_active",
        description="Specifies how the evaluator selects projects to work on.             One of: disabled, all_active, list",
        title="Project Assignment Mode",
    )
    context: Optional[Context] = Field(
        {"custom": {}}, description="Runtime properties to pass to executed tasks."
    )


class EvaluatorConfigurationUpdate(BaseModel):
    id: Optional[str] = Field(None, description="Unique database ID (read-only).", title="Id")
    evaluator_id: Optional[str] = Field(
        None, description="ID of the parent evaluator (read-only).", title="Evaluator Id"
    )
    last_modified: Optional[datetime] = Field(
        None, description="Last modified time", title="Last Modified"
    )
    working_directory: Optional[str] = Field(None, title="Working Directory")
    local_file_cache_max_size: Optional[int] = Field(
        None, description="Maximum allowed cache size in bytes.", title="Local File Cache Max Size"
    )
    max_num_parallel_tasks: Optional[int] = Field(None, title="Max Num Parallel Tasks")
    task_directory_cleanup: Optional[TaskDirectoryCleanupEnum] = Field(
        None, title="Task Directory Cleanup"
    )
    resources: Optional[EvaluatorResources] = {"custom": {}}
    name: Optional[str] = Field(
        None,
        description="Update the name of the evaluator (updating the registration).",
        title="Name",
    )
    loop_interval: Optional[float] = Field(
        None,
        description="Number of seconds between each iteration of the evaluator's main loop.",
        title="Loop Interval",
    )
    local_file_cache: Optional[bool] = Field(
        None,
        description="Whether to configure a local file cache in the file tool.",
        title="Local File Cache",
    )
    applications: Optional[List[ApplicationInfo]] = Field(
        [], description="List of available applications.", title="Applications"
    )
    project_list: Optional[List[str]] = Field(
        None,
        description="IDs of projects that the evaluator should work on in order.",
        title="Project List",
    )
    project_assignment_mode: Optional[str] = Field(
        None,
        description="How the evaluator selects projects to work on. Options are: disabled, all_active, list.",
        title="Project Assignment Mode",
    )
    context: Optional[ContextUpdate] = Field(
        {"custom": {}}, description="Runtime properties to pass to executed tasks."
    )


class EvaluatorConfigurationUpdatesRequest(BaseModel):
    configuration_updates: List[EvaluatorConfigurationUpdate] = Field(
        ..., description="Configuration update details", title="Configuration Updates"
    )


class EvaluatorConfigurationUpdatesResponse(BaseModel):
    configuration_updates: List[EvaluatorConfigurationUpdate] = Field(
        ..., description="Configuration update details", title="Configuration Updates"
    )


class EvaluatorConfigurationsResponse(BaseModel):
    configurations: List[EvaluatorConfiguration] = Field(
        ..., description="Evaluator configurations", title="Configurations"
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing the base class for all client and server HPS-related errors."""

from requests.exceptions import RequestException


class HPSError(RequestException):
    """Provides the base class for all HPS-related errors.

    This class derives from the :class:`requests.exceptions.RequestException`
    base class.

    Example:
        >>> from ansys.hps.client import HPSError
        >>> from ansys.hps.client.jms import Client
        >>> try:
        >>>     client = Client(url="https://127.0.0.1:8443/hps/",
                                username="repuser",
                                password="wrong_psw")
        >>> except HPSError as e:
        >>>     print(e)
        401 Client Error: invalid_grant for: POST https://127.0.0.1:8443/hps/auth...
        Invalid user credentials
    """

    def __init__(self, *args, **kwargs):
        self.reason = kwargs.pop("reason", None)
        self.description = kwargs.pop("description", None)
        super(HPSError, self).__init__(*args, **kwargs)


class APIError(HPSError):
    """Provides server-side related errors."""

    def __init__(self, *args, **kwargs):
        super(APIError, self).__init__(*args, **kwargs)


class ClientError(HPSError):
    """Provides client-side related errors."""

    def __init__(self, *args, **kwargs):
        super(ClientError, self).__init__(*args, **kwargs)


def raise_for_status(response, *args, **kwargs):
    """Automatically checks HTTP errors.

    This method mimics the requests.Response.raise_for_status() method.
    """
    if 400 <= response.status_code < 600:

        r_content = {}
        try:
            r_content = response.json()
        except ValueError:
            pass

        reason = r_content.get("title", None)  # jms api
        if not reason:
            reason = r_content.get("error", None)  # auth api
        if not reason:
            reason = response.reason

        description = r_content.get("description", None)  # jms api
        if not description:
            description = r_content.get("error_description", None)  # auth api

        if 400 <= response.status_code < 500:
            error_msg = "%s Client Error: %s for: %s %s" % (
                response.status_code,
                reason,
                response.request.method,
                response.url,
            )
            if description:
                error_msg += f"\n{description}"
            raise ClientError(error_msg, reason=reason, description=description, response=response)
        elif 500 <= response.status_code < 600:
            error_msg = "%s Server Error: %s for: %s %s" % (
                response.status_code,
                reason,
                response.request.method,
                response.url,
            )
            if description:
                error_msg += f"\n{description}"
            raise APIError(error_msg, reason=reason, description=description, response=response)
    return response



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing operation schema."""
from marshmallow import fields

from ansys.hps.client.common import ObjectSchema


class OperationSchema(ObjectSchema):
    class Meta(ObjectSchema.Meta):
        pass

    name = fields.String(allow_none=True)
    target = fields.List(fields.String(), allow_none=True)
    finished = fields.Bool(allow_none=True)
    succeeded = fields.Bool(allow_none=True)

    progress = fields.Float(allow_none=True)
    status = fields.String(allow_none=True)
    result = fields.Dict(allow_none=True)
    messages = fields.List(fields.Dict(), allow_none=True)

    start_time = fields.DateTime(allow_none=True)
    end_time = fields.DateTime(allow_none=True)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the task resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.task import TaskSchema
from .task_definition import TaskDefinition


class Task(Object):
    """Provides the task resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    pending_time : datetime, optional
        Date and time that the task was set to pending.
    prolog_time : datetime, optional
        Date and time that the task was set to prolog.
    running_time : datetime, optional
        Date and time that the task was set to running.
    finished_time : datetime, optional
        Date and time that the task was completed.
    eval_status : str, optional
        Evaluation status.
    trial_number : int, optional
        Which attempt to execute the process step this task represents.
    elapsed_time : float, optional
        Number of seconds it took the evaluator to execute the task.
    task_definition_id : str
        ID of the :class:`TaskDefinition` instance that the task is linked to.
    task_definition_snapshot : TaskDefinition, optional
        Snapshot of the  :class:`TaskDefinition` instance that was created when the task status changed to prolog, before evaluation.
    executed_command : str, optional
    job_id : str
        ID of the :class:`Job` instance that the task is linked to.
    host_id : str, optional
        UUID of the :class:`Evaluator` instance that updated the task.
    input_file_ids : list[str]
        List of IDs of input files of the task.
    output_file_ids : list[str]
        List of IDs of output files of the task.
    monitored_file_ids : list[str]
        List of IDs of monitored files of the task.
    inherited_file_ids : list[str]
        List of IDs of inherited files of the task.
    owned_file_ids : list[str]
        List of IDs of owned files of the task.
    license_context_id : str, optional
        ID of the license context in use.
    custom_data : dict, optional
        Dictionary type field for storing custom data.
    working_directory : str, optional
        Working directory of the task.
    """

    class Meta:
        schema = TaskSchema
        rest_name = "tasks"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        pending_time: datetime = missing,
        prolog_time: datetime = missing,
        running_time: datetime = missing,
        finished_time: datetime = missing,
        eval_status: str = missing,
        trial_number: int = missing,
        elapsed_time: float = missing,
        task_definition_id: str = missing,
        task_definition_snapshot: TaskDefinition = missing,
        executed_command: str = missing,
        job_id: str = missing,
        host_id: str = missing,
        input_file_ids: List[str] = missing,
        output_file_ids: List[str] = missing,
        monitored_file_ids: List[str] = missing,
        inherited_file_ids: List[str] = missing,
        owned_file_ids: List[str] = missing,
        license_context_id: str = missing,
        custom_data: Dict = missing,
        working_directory: str = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.pending_time = pending_time
        self.prolog_time = prolog_time
        self.running_time = running_time
        self.finished_time = finished_time
        self.eval_status = eval_status
        self.trial_number = trial_number
        self.elapsed_time = elapsed_time
        self.task_definition_id = task_definition_id
        self.task_definition_snapshot = task_definition_snapshot
        self.executed_command = executed_command
        self.job_id = job_id
        self.host_id = host_id
        self.input_file_ids = input_file_ids
        self.output_file_ids = output_file_ids
        self.monitored_file_ids = monitored_file_ids
        self.inherited_file_ids = inherited_file_ids
        self.owned_file_ids = owned_file_ids
        self.license_context_id = license_context_id
        self.custom_data = custom_data
        self.working_directory = working_directory

        self.obj_type = self.__class__.__name__


TaskSchema.Meta.object_class = Task



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module processing class members for an object."""
import json
import logging

from marshmallow.utils import missing

log = logging.getLogger(__name__)


class Object(object):
    class Meta:
        schema = None  # To be set in derived classes
        rest_name = (
            None  # String used in REST URI's to access this resource, to be set in derived classes
        )

    def declared_fields(self):
        """
        Provides a helper function for retrieving fields to define as class members for an object.
        """
        fields = []
        for k, v in self.Meta.schema._declared_fields.items():
            field = k
            # Ensure that we use the attribute name if defined
            if getattr(v, "attribute", None) is not None:
                field = v.attribute
            fields.append(field)
        return fields

    def __init__(self, **kwargs):
        # obj_type in JSON equals class name in API
        self.obj_type = self.__class__.__name__

        # Instantiate class members for all fields of the corresponding schema
        for k in self.declared_fields():

            # If property k is provided as init parameter
            if k in kwargs.keys():
                setattr(self, k, kwargs[k])
            # Else we set it this value as missing.
            # That way marshmallow will ignore it on serialization
            elif not hasattr(self, k):
                setattr(self, k, missing)

    def __repr__(self):
        """Printable representation of the object."""
        return "%s(%s)" % (
            self.__class__.__name__,
            ",".join(["%s=%r" % (k, getattr(self, k)) for k in self.declared_fields()]),
        )

    def __eq__(self, other):
        """Compare instances of the object."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        for k in self.declared_fields():
            if not hasattr(other, k) or getattr(self, k, None) != getattr(other, k, None):
                return False
        return True

    def __str__(self):
        """String representation of the object."""
        # Ideally we'd simply do
        #   return json.dumps(self.Meta.schema(many=False).dump(self), indent=2)
        # However the schema.dump() function (rightfully) ignores fields marked as load_only.
        #
        # Therefore we have to manually iterate over all fields

        schema = self.Meta.schema(many=False)
        dict_repr = schema.dict_class()
        for attr_name, field_obj in schema.fields.items():
            value = missing
            try:
                value = field_obj.serialize(attr_name, self, accessor=schema.get_attribute)
            except:
                pass
            if value is missing:
                continue
            key = field_obj.data_key if field_obj.data_key is not None else attr_name
            dict_repr[key] = value

        return json.dumps(dict_repr, indent=2)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""PyHPS JMS resource subpackage."""
from .algorithm import Algorithm
from .file import File
from .fitness_definition import FitnessDefinition, FitnessTermDefinition
from .job import Job
from .job_definition import JobDefinition
from .license_context import LicenseContext
from .operation import Operation
from .parameter_definition import (
    BoolParameterDefinition,
    FloatParameterDefinition,
    IntParameterDefinition,
    ParameterDefinition,
    StringParameterDefinition,
)
from .parameter_mapping import ParameterMapping
from .permission import Permission
from .project import Project
from .selection import JobSelection
from .task import Task
from .task_definition import (
    HpcResources,
    Licensing,
    ResourceRequirements,
    Software,
    SuccessCriteria,
    TaskDefinition,
)
from .task_definition_template import (
    TaskDefinitionTemplate,
    TemplateInputFile,
    TemplateOutputFile,
    TemplateProperty,
    TemplateResourceRequirements,
)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module retrieving IDs and providing ID references."""
import logging

from marshmallow import fields

# from ..keys import OBJECT_ID_KEY

log = logging.getLogger(__name__)


def id_from_ref(ref):
    return ref
    # if ref is None:
    #     return None
    # ref_dict = ref.get(OBJECT_REF_KEY)
    # if ref_dict is None:
    #     return None
    # return ref_dict[OBJECT_ID_KEY]


def id_to_ref(cls_name, id):
    return id
    # if id is None:
    #     return None
    # return {OBJECT_REF_KEY : {
    #             OBJECT_TYPE_KEY : cls_name,
    #             OBJECT_ID_KEY : int(id)
    #         }
    #     }


def id_list_to_ref(cls_name, ids):
    return ids
    # return {OBJECT_REF_LIST_KEY : {
    #             OBJECT_TYPE_KEY : cls_name,
    #             OBJECT_ID_LIST_KEY : [int(v) for v in ids]
    #         }
    #     }


def id_list_from_ref(ref):
    return ref
    # if ref is None:
    #     return None
    # ref_dict = ref.get(OBJECT_REF_LIST_KEY)
    # if ref_dict is None:
    #     return None
    # return ref_dict[OBJECT_ID_LIST_KEY]


class IdReference(fields.Field):
    def __init__(self, referenced_class, *args, **kwargs):
        super(IdReference, self).__init__(*args, **kwargs)
        self.referenced_class = referenced_class

    def _deserialize(self, value, attr, data, **kwargs):
        return id_from_ref(value)

    def _serialize(self, value, attr, obj, **kwargs):
        return id_to_ref(self.referenced_class, value)

    # def _validate(self, value):
    #     if not isinstance(value, int):
    #         raise ValidationError("Not an object reference: %s" % value)


class IdReferenceList(fields.Field):
    def __init__(self, referenced_class, *args, **kwargs):
        super(IdReferenceList, self).__init__(*args, **kwargs)
        self.referenced_class = referenced_class

    def _deserialize(self, value, attr, data, **kwargs):
        return id_list_from_ref(value)

    def _serialize(self, value, attr, obj, **kwargs):
        return id_list_to_ref(self.referenced_class, value)

    # def _validate(self, value):
    #     if not isinstance(value, list):
    #         raise ValidationError("Not an object reference list: %s" % value)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing job selection schema."""

from marshmallow import fields

from ansys.hps.client.common import ObjectSchemaWithModificationInfo

from .object_reference import IdReference, IdReferenceList


class JobSelectionSchema(ObjectSchemaWithModificationInfo):
    class Meta(ObjectSchemaWithModificationInfo.Meta):
        pass

    name = fields.String(metadata={"description": "Name of the selection."})

    algorithm_id = IdReference(
        allow_none=True,
        attribute="algorithm_id",
        referenced_class="DesignExplorationAlgorithm",
        metadata={
            "description": "ID of the :class:`Algorithm` instance that the selection belongs to."
        },
    )
    object_ids = IdReferenceList(
        "Job", attribute="jobs", metadata={"description": "List of job IDs."}
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing software, HPC resource, resource requirement, success criteria,
licensing, and task definition schema."""

from marshmallow import fields

from ansys.hps.client.common import BaseSchema, ObjectSchemaWithModificationInfo, RestrictedValue

from .object_reference import IdReference, IdReferenceList


class SoftwareSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        pass

    name = fields.String(metadata={"description": "Name of the app."})
    version = fields.String(allow_none=True, metadata={"description": "Version of the app."})


class HpcResourcesSchema(BaseSchema):
    class Meta:
        pass

    num_cores_per_node = fields.Int(
        allow_none=True, metadata={"description": "Number of cores per node."}
    )
    num_gpus_per_node = fields.Int(
        allow_none=True, metadata={"description": "Number of GPUs per node."}
    )
    exclusive = fields.Bool(
        allow_none=True,
        metadata={"description": "Whether a job can't share resources with other running jobs."},
    )
    queue = fields.Str(
        allow_none=True, metadata={"description": "Name of the job scheduler queue."}
    )


class ResourceRequirementsSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        pass

    platform = fields.String(
        allow_none=True,
        metadata={
            "description": "Basic platform information. Options are ``'linux'`` and ``'windows'``."
        },
    )
    memory = fields.Int(allow_none=True, metadata={"description": "Amount of RAM in bytes."})
    num_cores = fields.Float(allow_none=True, metadata={"description": "Number of cores."})
    disk_space = fields.Int(
        allow_none=True, metadata={"description": "Amount of disk space in bytes."}
    )
    distributed = fields.Bool(
        allow_none=True,
        metadata={"description": "Whether to enable distributed parallel processing."},
    )
    custom = fields.Dict(
        allow_none=True,
        keys=fields.Str(),
        values=RestrictedValue(),
        metadata={"description": "Dictionary of custom resource requirements."},
    )
    hpc_resources = fields.Nested(
        HpcResourcesSchema, allow_none=True, metadata={"description": "HPC resource requirements."}
    )


class SuccessCriteriaSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        pass

    return_code = fields.Int(
        allow_none=True,
        metadata={
            "description": "Process exit code that must be returned by the executed command."
        },
    )
    expressions = fields.List(
        fields.String(),
        allow_none=True,
        metadata={"description": "List of expressions to evaluate."},
    )

    required_output_file_ids = IdReferenceList(
        "File",
        attribute="required_output_file_ids",
        allow_none=True,
        metadata={"description": "List of IDs of the required output files."},
    )
    require_all_output_files = fields.Bool(
        allow_none=True, metadata={"description": "Whether to require all output files."}
    )

    required_output_parameter_ids = IdReferenceList(
        "ParameterDefinition",
        attribute="required_output_parameter_ids",
        allow_none=True,
        metadata={"description": "List of names of the required output parameters."},
    )
    require_all_output_parameters = fields.Bool(
        allow_none=True, metadata={"description": "Whether to require all output parameters."}
    )


class LicensingSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        pass

    enable_shared_licensing = fields.Bool(
        allow_none=True,
        metadata={
            "description": "Whether to enable shared licensing contexts for Ansys simulations."
        },
    )


class TaskDefinitionSchema(ObjectSchemaWithModificationInfo):
    class Meta(ObjectSchemaWithModificationInfo.Meta):
        pass

    name = fields.String(allow_none=True, metadata={"description": "Name."})

    execution_command = fields.String(
        allow_none=True,
        metadata={"description": "Command to execute. A command or execution script is required."},
    )
    use_execution_script = fields.Bool(
        allow_none=True,
        metadata={
            "description": "Whether to run the task with the execution script "
            "or the execution command."
        },
    )
    execution_script_id = IdReference(
        referenced_class="File",
        allow_none=True,
        metadata={"description": "Script to execute. A command or execution script is required."},
    )

    execution_level = fields.Int(metadata={"description": "Execution level for the task."})
    execution_context = fields.Dict(
        allow_none=True,
        metadata={"description": "Additional arguments to pass to the executing command."},
        keys=fields.Str(),
        values=RestrictedValue(),
    )
    environment = fields.Dict(
        allow_none=True,
        metadata={"description": "Environment variables to set for the executed process."},
        keys=fields.Str(),
        values=fields.Str(),
    )
    max_execution_time = fields.Float(
        allow_none=True, metadata={"description": "Maximum time in seconds for executing the task."}
    )
    num_trials = fields.Int(
        allow_none=True,
        metadata={"description": "Maximum number of attempts for executing the task."},
    )
    store_output = fields.Bool(
        allow_none=True,
        metadata={"description": "Whether to store the standard output of the task."},
    )

    input_file_ids = IdReferenceList(
        referenced_class="File",
        attribute="input_file_ids",
        metadata={"description": "List of IDs of input files."},
    )
    output_file_ids = IdReferenceList(
        referenced_class="File",
        attribute="output_file_ids",
        metadata={"description": "List of IDs of output files."},
    )

    success_criteria = fields.Nested(
        SuccessCriteriaSchema,
        allow_none=True,
    )
    licensing = fields.Nested(
        LicensingSchema,
        allow_none=True,
        metadata={"description": ":class:`Licensing` object."},
    )

    software_requirements = fields.Nested(
        SoftwareSchema,
        many=True,
        allow_none=True,
        metadata={"description": "List of :class:`Software` objects."},
    )
    resource_requirements = fields.Nested(
        ResourceRequirementsSchema,
        allow_none=True,
        metadata={"description": ":class:`ResourceRequirements` object."},
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the permission resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.permission import PermissionSchema


class Permission(Object):
    """Provides the permission resource.

    Parameters
    ----------
    permission_type : str
        Permission type. Options are ``'anyone'``, ``'group'``, and ``'user'``.
    value_id : str, optional
        ID of a user or group.
    value_name : str, optional
    role : str
        Role of the user. Options are ``'admin'``, ``'reader'``, and ``'writer'``.
    """

    class Meta:
        schema = PermissionSchema
        rest_name = "permissions"

    def __init__(
        self,
        permission_type: str = missing,
        value_id: str = missing,
        value_name: str = missing,
        role: str = missing,
        **kwargs,
    ):
        self.permission_type = permission_type
        self.value_id = value_id
        self.value_name = value_name
        self.role = role

        self.obj_type = self.__class__.__name__


PermissionSchema.Meta.object_class = Permission



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing base schemas and object schemas with and without modification information."""

from marshmallow import INCLUDE, Schema, fields, post_load


class BaseSchema(Schema):
    """Base schema class."""

    class Meta:
        ordered = True
        unknown = INCLUDE
        object_class = None  # To be set in derived objects

    @post_load
    def make_object(self, data, **kwargs):
        """Make object for base schema."""
        return self.Meta.object_class(**data)


class ObjectSchema(BaseSchema):
    """Create object schema with ID."""

    id = fields.String(
        allow_none=True,
        attribute="id",
        metadata={
            "description": "Unique ID to access the resource, generated "
            "internally by the server on creation."
        },
    )


class ObjectSchemaWithModificationInfo(ObjectSchema):
    """Object schema with creation & modification times, and created & modified by fields."""

    creation_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={
            "description": "Date and time that the resource was created.",
        },
    )
    modification_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={
            "description": "Date and time that the resource was last modified.",
        },
    )

    created_by = fields.String(
        allow_none=True,
        load_only=True,
        metadata={
            "description": "ID of the user who created the object.",
        },
    )
    modified_by = fields.String(
        allow_none=True,
        load_only=True,
        metadata={
            "description": "ID of the user who last modified the object.",
        },
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module exposing the project endpoints of the JMS."""
import json
import logging
import os
from pathlib import Path
from typing import Callable, List, Type, Union
from warnings import warn

import requests

from ansys.hps.client.client import Client
from ansys.hps.client.common import Object
from ansys.hps.client.exceptions import HPSError
from ansys.hps.client.jms.resource import (
    Algorithm,
    File,
    Job,
    JobDefinition,
    JobSelection,
    LicenseContext,
    ParameterDefinition,
    ParameterMapping,
    Permission,
    Project,
    Task,
    TaskDefinition,
)

from .base import create_objects, delete_objects, get_objects, update_objects
from .jms_api import JmsApi, _copy_objects

log = logging.getLogger(__name__)


class ProjectApi:
    """Exposes the project endpoints of the JMS.

    Parameters
    ----------
    client : Client
        HPS client object.
    project_id : str
        ID of the project.

    Examples
    --------

    >>> from ansys.hps.client import Client
    >>> from ansys.hps.client.jms import JmsApi, Project, ProjectApi
    >>> cl = Client(
    ...     url="https://127.0.0.1:8443/hps", username="repuser", password="repuser"
    ... )
    >>> project = Project(name="Example project")
    >>> print(project)
    {
        "name": "Example project"
    }
    >>> jms_api = JmsApi(cl)
    >>> project = jms_api.create_project(project)
    >>> print(project)
    {
        "id": "02qtyJfpfAQ0fr3zkoIAfC",
        "name": "Example project",
        "active": false,
        "priority": 1,
        "creation_time": ...
        ...
    }
    >>> project_api = ProjectApi(cl, project.id)
    >>> print(project_api)
    'https://127.0.0.1:8443/hps/jms/api/v1/projects/02qtyJfpfAQ0fr3zkoIAfC'
    >>> jobs = project_api.get_jobs()

    """

    def __init__(self, client: Client, project_id: str):
        """Initialize project API."""
        self.client = client
        self.project_id = project_id
        self._fs_url = None
        self._fs_project_id = None

    @property
    def jms_api_url(self) -> str:
        """Get the JMS API URL."""
        return f"{self.client.url}/jms/api/v1"

    @property
    def url(self) -> str:
        """URL of the API."""
        return f"{self.jms_api_url}/projects/{self.project_id}"

    @property
    def fs_url(self) -> str:
        """URL of the file storage gateway."""
        if self._fs_url is None:
            self._fs_url = JmsApi(self.client).fs_url
        return self._fs_url

    @property
    def fs_bucket_url(self) -> str:
        """URL of the project's bucket in the file storage gateway."""
        return f"{self.fs_url}/{self.project_id}"

    ################################################################
    # Project operations (copy, archive)
    def copy_project(self, wait: bool = True) -> str:
        """Duplicate the project."""
        r = copy_projects(self, [self.project_id], wait)
        if wait:
            return r[0]
        else:
            return r

    def archive_project(self, path: str, include_job_files: bool = True):
        """Archive a project and save it to disk.

        Parameters
        ----------
        path : str
            Path for saving the archive locally.
        include_job_files : bool, optional
            Whether to include job files in the archive. The default is ``True``.

        Returns
        -------
        str
            Path to the archive.
        """
        return archive_project(self, path, include_job_files)

    ################################################################
    # Files
    def get_files(self, as_objects=True, content=False, **query_params) -> List[File]:
        """
        Get a list of file resources, optionally filtered by query parameters.

        If ``content=True``, each file's content is also downloaded and stored in memory
        as the :attr:`ansys.hps.client.jms.File.content` attribute.
        """
        return get_files(self, as_objects=as_objects, content=content, **query_params)

    def create_files(self, files: List[File], as_objects=True) -> List[File]:
        """Create a list of files."""
        return create_files(self, files, as_objects=as_objects)

    def update_files(self, files: List[File], as_objects=True):
        """Update files."""
        return update_files(self, files, as_objects=as_objects)

    def delete_files(self, files: List[File]):
        """Delete files."""
        return self._delete_objects(files, File)

    def download_file(
        self,
        file: File,
        target_path: str,
        stream: bool = True,
        progress_handler: Callable[[int], None] = None,
    ) -> str:
        """
        Download file content and save it to disk.

        If ``stream=True``, data is retrieved in chunks, which avoids storing the entire content
        in memory.
        """
        return _download_file(self, file, target_path, progress_handler, stream)

    ################################################################
    # Parameter definitions
    def get_parameter_definitions(
        self, as_objects=True, **query_params
    ) -> List[ParameterDefinition]:
        """Get a list of parameter definitions."""
        return self._get_objects(ParameterDefinition, as_objects, **query_params)

    def create_parameter_definitions(
        self, parameter_definitions: List[ParameterDefinition], as_objects=True
    ) -> List[ParameterDefinition]:
        """Create a list of parameter definitions."""
        return self._create_objects(parameter_definitions, ParameterDefinition, as_objects)

    def update_parameter_definitions(
        self, parameter_definitions: List[ParameterDefinition], as_objects=True
    ) -> List[ParameterDefinition]:
        """Update a list of parameter definitions."""
        return self._update_objects(parameter_definitions, ParameterDefinition, as_objects)

    def delete_parameter_definitions(self, parameter_definitions: List[ParameterDefinition]):
        """Delete a list of parameter definitions."""
        return self._delete_objects(parameter_definitions, ParameterDefinition)

    ################################################################
    # Parameter mappings
    def get_parameter_mappings(self, as_objects=True, **query_params) -> List[ParameterMapping]:
        """Get a list of parameter mappings."""
        return self._get_objects(ParameterMapping, as_objects=as_objects, **query_params)

    def create_parameter_mappings(
        self, parameter_mappings: List[ParameterMapping], as_objects=True
    ) -> List[ParameterMapping]:
        """Get a list of created parameter mappings."""
        return self._create_objects(parameter_mappings, ParameterMapping, as_objects=as_objects)

    def update_parameter_mappings(
        self, parameter_mappings: List[ParameterMapping], as_objects=True
    ) -> List[ParameterMapping]:
        """Get a list of updated parameter mappings."""
        return self._update_objects(parameter_mappings, ParameterMapping, as_objects=as_objects)

    def delete_parameter_mappings(self, parameter_mappings: List[ParameterMapping]):
        """Delete a list of parameter mappings."""
        return self._delete_objects(parameter_mappings, ParameterMapping)

    ################################################################
    # Task definitions
    def get_task_definitions(self, as_objects=True, **query_params) -> List[TaskDefinition]:
        """Get a list of task definitions."""
        return self._get_objects(TaskDefinition, as_objects=as_objects, **query_params)

    def create_task_definitions(
        self, task_definitions: List[TaskDefinition], as_objects=True
    ) -> List[TaskDefinition]:
        """Create a list of task definitions."""
        return self._create_objects(task_definitions, TaskDefinition, as_objects=as_objects)

    def update_task_definitions(
        self, task_definitions: List[TaskDefinition], as_objects=True
    ) -> List[TaskDefinition]:
        """Update a list of task definitions."""
        return self._update_objects(task_definitions, TaskDefinition, as_objects=as_objects)

    def delete_task_definitions(self, task_definitions: List[TaskDefinition]):
        """Delete a list of task definitions."""
        return self._delete_objects(task_definitions, TaskDefinition)

    def copy_task_definitions(
        self, task_definitions: List[TaskDefinition], wait: bool = True
    ) -> Union[str, List[str]]:
        """Create task definitions by copying existing task definitions.

        Parameters
        ----------
        task_definitions : List[TaskDefinition]
            List of task definitions. Note that only the ``id`` field of the
            ``TaskDefinition`` objects must be filled. Other fields can be empty.
        wait : bool
            Whether to wait for the copy to complete. The default is ``True``.

        Returns
        -------
        Union[List[str], str]
            If ``wait=True``, returns the list of newly created task definition IDs.
            If ``wait=False``, returns an operation ID that can be used to
            track progress.
        """
        return _copy_objects(self.client, self.url, task_definitions, wait=wait)

    ################################################################
    # Job definitions
    def get_job_definitions(self, as_objects=True, **query_params) -> List[JobDefinition]:
        """Get a list of job definitions."""
        return self._get_objects(JobDefinition, as_objects=as_objects, **query_params)

    def create_job_definitions(
        self, job_definitions: List[JobDefinition], as_objects=True
    ) -> List[JobDefinition]:
        """Create a list of job definitions."""
        return self._create_objects(job_definitions, JobDefinition, as_objects=as_objects)

    def update_job_definitions(
        self, job_definitions: List[JobDefinition], as_objects=True
    ) -> List[JobDefinition]:
        """Update a list of job definitions."""
        return self._update_objects(job_definitions, JobDefinition, as_objects=as_objects)

    def delete_job_definitions(self, job_definitions: List[JobDefinition]):
        """Delete a list of job definitions."""
        return self._delete_objects(job_definitions, JobDefinition)

    def copy_job_definitions(
        self, job_definitions: List[JobDefinition], wait: bool = True
    ) -> Union[str, List[str]]:
        """Create job definitions by copying existing job definitions.

        Parameters
        ----------
        job_definitions : List[JobDefinition]
            List of job definitions. Note that only the ``id`` field of the
            ``JobDefinition`` objects must be filled. The other fields can be empty.
        wait : bool
            Whether to wait for the copy to complete. The default is ``True``.

        Returns
        -------
        Union[List[str], str]
            If ``wait=True``, returns the list of newly created job definition IDs.
            If ``wait=False``, returns an operation ID that can be used to
            track progress.
        """
        return _copy_objects(self.client, self.url, job_definitions, wait=wait)

    ################################################################
    # Jobs
    def get_jobs(self, as_objects=True, **query_params) -> List[Job]:
        """Get a list of jobs."""
        return self._get_objects(Job, as_objects=as_objects, **query_params)

    def create_jobs(self, jobs: List[Job], as_objects=True) -> List[Job]:
        """Create jobs.

        Parameters
        ----------
        jobs : list of :class:`ansys.hps.client.jms.Job`
            List of jobs.
        as_objects : bool, optional
            Whether to return jobs as objects. The default is ``True``. If
            ``False``, jobs are returned as dictionaries.

        Returns
        -------
        list
            List of :class:`ansys.hps.client.jms.Job` objects if ``as_objects=True`` or
            a list of dictionaries if ``as_objects=False``.
        """
        return self._create_objects(jobs, Job, as_objects=as_objects)

    def copy_jobs(self, jobs: List[Job], wait: bool = True) -> Union[str, List[str]]:
        """Create jobs by copying existing jobs.

        Parameters
        ----------
        jobs : List[Job]
            List of jobs. Note that only the ``id`` field of the
            ``Job`` objects must be filled. The other fields can be empty.
        wait : bool, optional
            Whether to wait for the copy to complete. The default is ``True``.

        Returns
        -------
        Union[List[str], str]
            If ``wait=True``, returns the list of newly created job IDs.
            If ``wait=False``, returns an operation ID that can be used to
            track progress.
        """
        return _copy_objects(self.client, self.url, jobs, wait=wait)

    def update_jobs(self, jobs: List[Job], as_objects=True) -> List[Job]:
        """Update jobs.

        Parameters
        ----------
        jobs : list of :class:`ansys.hps.client.jms.Job`)
            List of jobs.
        as_objects : bool, optional
            Whether to return jobs as objects. The default is ``True``.
            If ``False``, jobs are returned as dictionaries.

        Returns
        -------
        list
            List of :class:`ansys.hps.client.jms.Job` objects if ``as_objects=True`` or a list of
            dictionaries if ``as_objects=False``.
        """
        return self._update_objects(jobs, Job, as_objects=as_objects)

    def delete_jobs(self, jobs: List[Job]):
        """Delete jobs.

        Parameters
        ----------
        jobs : list of :class:`ansys.hps.client.jms.Job`
            List of `jobs. Note that only the ``id`` field of the ``Job`` objects must be filled.
            The other fields can be empty.

        Example:

            >>> jobs_to_delete = []
            >>> for id in [1,2,39,44]:
            >>>    jobs_to_delete.append(Job(id=id))
            >>> project_api.delete_jobs(jobs_to_delete)

        """
        return self._delete_objects(jobs, Job)

    def sync_jobs(self, jobs: List[Job]):
        """Sync a list of jobs."""
        return sync_jobs(self, jobs)

    def _sync_jobs(self, jobs: List[Job]):
        """Deprecated function that syncs a list of jobs."""
        msg = (
            "'ProjectApi._sync_jobs' is deprecated and is to be removed soon. "
            "Use 'ProjectApi.sync_jobs' instead."
        )
        warn(msg, DeprecationWarning)
        log.warning(msg)
        return self.sync_jobs(jobs)

    ################################################################
    # Tasks
    def get_tasks(self, as_objects=True, **query_params) -> List[Task]:
        """Get a list of tasks."""
        return self._get_objects(Task, as_objects=as_objects, **query_params)

    def update_tasks(self, tasks: List[Task], as_objects=True) -> List[Task]:
        """Update a list of tasks."""
        return self._update_objects(tasks, Task, as_objects=as_objects)

    ################################################################
    # Selections
    def get_job_selections(self, as_objects=True, **query_params) -> List[JobSelection]:
        """Get a list of job selections."""
        return self._get_objects(JobSelection, as_objects=as_objects, **query_params)

    def create_job_selections(
        self, selections: List[JobSelection], as_objects=True
    ) -> List[JobSelection]:
        """Create a list of job selections."""
        return self._create_objects(selections, JobSelection, as_objects=as_objects)

    def update_job_selections(
        self, selections: List[JobSelection], as_objects=True
    ) -> List[JobSelection]:
        """Update a list of job selections."""
        return self._update_objects(selections, JobSelection, as_objects=as_objects)

    def delete_job_selections(self, selections: List[JobSelection]):
        """Delete a list of job selections."""
        return self._delete_objects(selections, JobSelection)

    ################################################################
    # Algorithms
    def get_algorithms(self, as_objects=True, **query_params) -> List[Algorithm]:
        """Get a list of algorithms."""
        return self._get_objects(Algorithm, as_objects=as_objects, **query_params)

    def create_algorithms(self, algorithms: List[Algorithm], as_objects=True) -> List[Algorithm]:
        """Create a list of algorithms."""
        return self._create_objects(algorithms, Algorithm, as_objects=as_objects)

    def update_algorithms(self, algorithms: List[Algorithm], as_objects=True) -> List[Algorithm]:
        """Update a list of algorithms."""
        return self._update_objects(algorithms, Algorithm, as_objects=as_objects)

    def delete_algorithms(self, algorithms: List[Algorithm]):
        """Delete a list of algorithms."""
        return self._delete_objects(algorithms, Algorithm)

    ################################################################
    # Permissions
    def get_permissions(self, as_objects=True) -> List[Permission]:
        """Get a list of permissions."""
        return self._get_objects(Permission, as_objects=as_objects, fields=None)

    def update_permissions(self, permissions: List[Permission], as_objects=True):
        """Update a list of permissions."""
        return self._update_objects(permissions, Permission, as_objects=as_objects)

    ################################################################
    # License contexts
    def get_license_contexts(self, as_objects=True, **query_params) -> List[LicenseContext]:
        """Get a list of license contexts."""
        return self._get_objects(self, LicenseContext, as_objects=as_objects, **query_params)

    def create_license_contexts(self, as_objects=True) -> List[LicenseContext]:
        """Create a list of license contexts."""
        rest_name = LicenseContext.Meta.rest_name
        url = f"{self.jms_api_url}/projects/{self.project_id}/{rest_name}"
        r = self.client.session.post(f"{url}")
        data = r.json()[rest_name]
        if not as_objects:
            return data
        schema = LicenseContext.Meta.schema(many=True)
        objects = schema.load(data)
        return objects

    def update_license_contexts(self, license_contexts, as_objects=True) -> List[LicenseContext]:
        """Update a list of license contexts."""
        return self._update_objects(self, license_contexts, LicenseContext, as_objects=as_objects)

    def delete_license_contexts(self):
        """Delete license contexts."""
        rest_name = LicenseContext.Meta.rest_name
        url = f"{self.jms_api_url}/projects/{self.id}/{rest_name}"
        r = self.client.session.delete(url)

    ################################################################
    def copy_default_execution_script(self, filename: str) -> File:
        """Copy a default execution script to the current project.

        Example:

            >>> file = project_api.copy_default_execution_script("exec_mapdl.py")

        """

        # create file resource
        name = os.path.splitext(filename)[0]
        file = File(name=name, evaluation_path=filename, type="application/x-python-code")
        file = self.create_files([file])[0]

        # query location of default execution scripts from server
        jms_api = JmsApi(self.client)
        info = jms_api.get_api_info()
        execution_script_default_bucket = info["settings"]["execution_script_default_bucket"]

        # server side copy of the file to project bucket
        checksum = _fs_copy_file(
            self.client.session,
            self.fs_url,
            execution_script_default_bucket,
            filename,
            self.project_id,
            file.storage_id,
        )

        # update file resource
        file.hash = checksum
        return self.update_files([file])[0]

    ################################################################
    def _get_objects(self, obj_type: Object, as_objects=True, **query_params):
        """Get objects."""
        return get_objects(self.client.session, self.url, obj_type, as_objects, **query_params)

    def _create_objects(
        self, objects: List[Object], obj_type: Type[Object], as_objects=True, **query_params
    ):
        """Create objects."""
        return create_objects(
            self.client.session, self.url, objects, obj_type, as_objects, **query_params
        )

    def _update_objects(
        self, objects: List[Object], obj_type: Type[Object], as_objects=True, **query_params
    ):
        """Update objects."""
        return update_objects(
            self.client.session, self.url, objects, obj_type, as_objects, **query_params
        )

    def _delete_objects(self, objects: List[Object], obj_type: Type[Object]):
        """Delete objects."""
        delete_objects(self.client.session, self.url, objects, obj_type)


def _download_files(project_api: ProjectApi, files: List[File]):
    """
    Download files directly using the fs REST gateway.

    This is a temporary implementation for downloading files. It is to be
    replaced with direct ansft calls, when it is available as a Python package.
    """

    for f in files:
        if getattr(f, "hash", None) is not None:
            r = project_api.client.session.get(f"{project_api.fs_bucket_url}/{f.storage_id}")
            f.content = r.content
            f.content_type = r.headers["Content-Type"]


def get_files(project_api: ProjectApi, as_objects=True, content=False, **query_params):
    """Get files for the project API."""
    files = get_objects(
        project_api.client.session, project_api.url, File, as_objects=as_objects, **query_params
    )
    if content:
        _download_files(project_api, files)
    return files


def _upload_files(project_api: ProjectApi, files):
    """
    Uploads files directly using the fs REST gateway.

    This is a temporary implementation for uploading files. It is to be
    replaced with direct ansft calls, when it is available as a Python package.
    """
    fs_headers = {"content-type": "application/octet-stream"}

    for f in files:
        if getattr(f, "src", None) is None:
            continue

        is_file = isinstance(f.src, str) and os.path.exists(f.src)
        content = f.src
        if is_file:
            content = open(f.src, "rb")

        r = project_api.client.session.post(
            f"{project_api.fs_bucket_url}/{f.storage_id}",
            data=content,
            headers=fs_headers,
        )
        f.hash = r.json()["checksum"]
        f.size = r.request.headers.get("Content-Length", None)

        if is_file:
            content.close()


def create_files(project_api: ProjectApi, files, as_objects=True) -> List[File]:
    """Create a list of files."""
    # (1) Create file resources in JMS
    created_files = create_objects(
        project_api.client.session, project_api.url, files, File, as_objects=as_objects
    )

    # (2) Check if there are src properties, files to upload
    num_uploads = 0
    for f, cf in zip(files, created_files):
        if getattr(f, "src", None) is not None:
            cf.src = f.src
            num_uploads += 1

    if num_uploads > 0:

        # (3) Upload file contents
        _upload_files(project_api, created_files)

        # (4) Update corresponding file resources in JMS with hashes of uploaded files
        created_files = update_objects(
            project_api.client.session, project_api.url, created_files, File, as_objects=as_objects
        )

    return created_files


def update_files(project_api: ProjectApi, files: List[File], as_objects=True) -> List[File]:
    """Update a list of files."""
    # Upload files first if there are any src parameters
    _upload_files(project_api, files)
    # Update file resources in JMS
    return update_objects(
        project_api.client.session, project_api.url, files, File, as_objects=as_objects
    )


def _download_file(
    project_api: ProjectApi,
    file: File,
    target_path: str,
    progress_handler: Callable[[int], None] = None,
    stream: bool = True,
) -> str:
    """Download a file."""
    if getattr(file, "hash", None) is None:
        log.warning(f"No hash found for file {file.name}.")

    download_link = f"{project_api.fs_bucket_url}/{file.storage_id}"
    download_path = os.path.join(target_path, file.evaluation_path)
    Path(download_path).parent.mkdir(parents=True, exist_ok=True)

    with (
        project_api.client.session.get(download_link, stream=stream) as r,
        open(download_path, "wb") as f,
    ):
        for chunk in r.iter_content(chunk_size=None):
            f.write(chunk)
            if progress_handler is not None:
                progress_handler(len(chunk))

    return download_path


def copy_projects(
    project_api: ProjectApi, project_source_ids: List[str], wait: bool = True
) -> Union[str, List[str]]:
    """Copy projects."""
    return _copy_objects(
        project_api.client,
        project_api.jms_api_url,
        [Project(id=id) for id in project_source_ids],
        wait=wait,
    )


def archive_project(project_api: ProjectApi, target_path, include_job_files=True) -> str:
    """Archive projects."""
    # PUT archive request
    url = f"{project_api.url}/archive"
    query_params = {}
    if not include_job_files:
        query_params["download_files"] = "configuration"

    r = project_api.client.session.put(url, params=query_params)

    # Monitor archive operation
    operation_location = r.headers["location"]
    log.debug(f"Operation location: {operation_location}")
    operation_id = operation_location.rsplit("/", 1)[-1]

    jms_api = JmsApi(project_api.client)
    op = jms_api.monitor_operation(operation_id)

    if not op.succeeded:
        raise HPSError(f"Failed to archive project {project_api.project_id}.\n{op}")

    download_link = op.result["backend_path"]

    # Download archive
    download_link = download_link.replace("ansfs://", project_api.fs_url + "/")
    log.info(f"Project archive download link: {download_link}")

    if not os.path.isdir(target_path):
        raise HPSError(f"Project archive: target path does not exist {target_path}")

    file_path = os.path.join(target_path, download_link.rsplit("/")[-1])
    log.info(f"Download archive to {file_path}")

    with project_api.client.session.get(download_link, stream=True) as r:
        with open(file_path, "wb") as f:
            for chunk in r.iter_content(chunk_size=1024 * 1024):
                if chunk:
                    f.write(chunk)

    log.info(f"Done saving project archive to disk.")
    return file_path


def copy_jobs(project_api: ProjectApi, jobs: List[Job], as_objects=True, **query_params):
    """Create jobs by copying existing jobs."""

    ids = _copy_objects(client=project_api.client, api_url=project_api.url, objects=jobs, wait=True)
    return ids


def sync_jobs(project_api: ProjectApi, jobs: List[Job]):
    """Sync jobs."""
    url = f"{project_api.url}/jobs:sync"  # noqa: E231
    json_data = json.dumps({"job_ids": [obj.id for obj in jobs]})
    r = project_api.client.session.put(f"{url}", data=json_data)


def _fs_copy_file(
    session: requests.Session,
    fs_url: str,
    source_bucket: str,
    source_name: str,
    destination_bucket: str,
    destination_name: str,
) -> str:
    """Copy files with the fs REST gateway."""
    json_data = json.dumps(
        {"destination": f"ansfs://{destination_bucket}/{destination_name}"}  # noqa: E231
    )
    r = session.post(
        url=f"{fs_url}/{source_bucket}/{source_name}:copy", data=json_data  # noqa: E231
    )
    return r.json()["checksum"]
